# frozen_string_literal: true

require 'apple_core/refinements/deep_dup'
require 'apiphobic/errors/unpermitted_inclusions'
require 'apiphobic/errors/unpermitted_sorts'

# rubocop:disable Metrics/ClassLength
module  Apiphobic
module  Authorization
module  Authorizers
class   Parameters
  using ::AppleCore::Refinements::DeepDup

  attr_accessor :action,
                :audience,
                :parameters,
                :token

  attr_writer   :authorized_attributes,
                :authorized_filters,
                :authorized_inclusions,
                :authorized_relationships,
                :authorized_sorts,
                :ignored_attributes

  # rubocop:disable Metrics/ParameterLists
  def initialize(action:, token:, audience:, issuer:, parameters:, **other)
    self.action     = action
    self.audience   = audience
    self.parameters = parameters.deep_dup
    self.token      = token

    other.each do |name, value|
      public_send("#{name}=", value)
    end
  end
  # rubocop:enable Metrics/ParameterLists

  alias user audience

  def authorized_parameters
    @authorized_parameters || [
                                :include,
                                :sort,
                                {
                                  name:                :data,
                                  authorization_value: [
                                                         :type,
                                                         :id,
                                                         {
                                                           attributes:    nil,
                                                           relationships: nil,
                                                         },
                                                       ],
                                },
                                {
                                  name:                :filter,
                                  authorization_value: [{}],
                                },
                                {
                                  name:                :page,
                                  authorization_value: %i{
                                                         number
                                                         size
                                                         offset
                                                         limit
                                                         cursor
                                                       },
                                },
                              ]
  end

  def authorized_attributes
    @authorized_attributes || []
  end

  def authorized_filters
    @authorized_filters || []
  end

  def authorized_inclusions
    @authorized_inclusions || []
  end

  def authorized_relationships
    @authorized_relationships || []
  end

  def authorized_sorts
    @authorized_sorts || []
  end

  def ignored_attributes
    @ignored_attributes || []
  end

  # rubocop:disable Metrics/AbcSize, Metrics/CyclomaticComplexity
  def call
    sliced_parameters = authorized_parameters.map { |v| v.is_a?(::Hash) ? v[:name] : v }

    parameters.slice!(*sliced_parameters)

    authorized_parameters.each do |parameter|
      parameter = { name: parameter } unless parameter.is_a?(::Hash)

      authorize_parameter(value:                    parameters[parameter],
                          authorization_parameters: authorization,
                          raw_parameters:           parameters,
                          **parameter)
    end

    authorized_attributes.each do |attribute|
      attribute = { name: attribute } unless attribute.is_a?(::Hash)

      authorize_attribute(**attribute)
    end

    ignored_attributes.each do |attribute|
      attribute = { name: attribute } unless attribute.is_a?(::Hash)

      ignore_attribute(**attribute)
    end

    authorized_filters.each do |filter|
      filter = { name: filter } unless filter.is_a?(::Hash)

      authorize_filter(**filter)
    end

    authorized_relationships.each do |name|
      name = { name: name } unless name.is_a?(::Hash)

      authorize_relationship(**name)
    end

    authorize_inclusions(names: authorized_inclusions)
    authorize_sorts(names: authorized_sorts)

    parameters.permit(*authorization)
  end
  # rubocop:enable Metrics/AbcSize, Metrics/CyclomaticComplexity

  private

  def authorization
    @authorization ||= [{}]
  end

  def authorize_attribute(**args)
    authorize_parameter(
      value:                    raw_parameter_attribute_value(args[:name]),
      authorization_parameters: authorized_parameter_attributes,
      raw_parameters:           raw_parameter_attributes,
      **args,
    )
  end

  def authorize_filter(**args)
    authorize_parameter(
      value:                    raw_parameter_filter_value(args[:name]),
      authorization_parameters: authorized_parameter_filters,
      raw_parameters:           raw_parameter_filters,
      **args,
    )
  end

  def authorize_inclusions(names:)
    return if names.empty?

    all_requested_inclusions_authorized = raw_parameter_inclusions
                                            .to_s
                                            .split(',')
                                            .all? do |inclusion|
                                              names.map(&:to_s).include?(inclusion.to_s)
                                            end

    fail Errors::UnpermittedInclusions.new(inclusions: raw_parameter_inclusions) \
      unless all_requested_inclusions_authorized
  end

  def authorize_sorts(names:)
    return if names.empty?

    all_requested_sorts_authorized = raw_parameter_sorts
                                       .to_s
                                       .delete('-')
                                       .split(',')
                                       .all? do |sort|
                                         names.map(&:to_s).include?(sort.to_s)
                                       end

    fail Errors::UnpermittedSorts.new(sorts: raw_parameter_sorts) \
      unless all_requested_sorts_authorized
  end

  # rubocop:disable Metrics/ParameterLists
  def authorize_parameter(name:,
                          value:,
                          authorization_parameters:,
                          authorization_value: nil,
                          raw_parameters:,
                          allowed_values: nil,
                          override: {})

    value = override_parameter(name:     name,
                               value:    value,
                               hash:     raw_parameters,
                               override: override)

    return unless allowed_values.nil? || allowed_values.include?(value)

    if authorization_value
      authorization_parameters[0][name] = authorization_value
    elsif value.class == ::Array
      authorization_parameters[0][name] = []
    else
      authorization_parameters << name
    end
  end
  # rubocop:enable Metrics/ParameterLists

  def authorize_relationship(name:, embedded_attributes: [])
    relationship_data = raw_parameter_relationship_data(name)
    to_many_relation  = relationship_data.is_a?(::Array)
    first_relation    = if to_many_relation
                          relationship_data[0]
                        else
                          relationship_data || {}
                        end
    embedded_relation = first_relation[:attributes]

    authorized_parameter_relationships[name] = if relationship_data.nil?
                                                 %i{data}
                                               elsif embedded_relation
                                                 {
                                                   data: [
                                                           :id,
                                                           :type,
                                                           {
                                                             attributes: %i{__id__} +
                                                             embedded_attributes,
                                                           },
                                                         ],
                                                 }
                                               else
                                                 { data: %i{type id} }
                                               end
  end

  def ignore_attribute(name:)
    key_match = ->(k, _v) { k.to_s == name.to_s }

    parameters[:data][:attributes]    = raw_parameter_attributes.delete_if(&key_match)
    parameters[:data][:relationships] = raw_parameter_relationships.delete_if(&key_match)
  end

  # rubocop:disable Metrics/CyclomaticComplexity, Metrics/PerceivedComplexity
  def override_parameter(name:, value:, hash:, override:)
    override = { with: nil, if_admin: false, if_absent: true, if_blank: true }
                 .merge(override)

    return value unless override[:with] &&
                        (!token.admin? || override[:if_admin]) &&
                        (hash.has_key?(name) || override[:if_absent]) &&
                        (!hash.has_key?(name) || !value.nil? || override[:if_blank])

    hash[name] = override[:with]

    override[:with]
  end
  # rubocop:enable Metrics/CyclomaticComplexity, Metrics/PerceivedComplexity

  def authorized_data_parameter
    authorization[0][:data][2]
  end

  def authorized_parameter_attributes
    authorized_data_parameter[:attributes] ||= [{}]
  end

  def authorized_parameter_relationships
    authorized_data_parameter[:relationships] ||= {}
  end

  def authorized_parameter_filters
    authorization[0][:filter] ||= [{}]
  end

  # rubocop:disable Layout/ExtraSpacing
  def raw_parameter_attributes
    @raw_parameter_attributes ||= begin
      parameters[:data]              ||= {}
      parameters[:data][:attributes] ||= {}

      parameters[:data][:attributes]
    end
  end
  # rubocop:enable Layout/ExtraSpacing

  def raw_parameter_filters
    @raw_parameter_filters ||= parameters[:filter] ||= {}
  end

  def raw_parameter_inclusions
    @raw_parameter_inclusions ||= parameters[:include]
  end

  def raw_parameter_relationships
    @raw_parameter_relationships        ||= begin
      parameters[:data]                 ||= {}
      parameters[:data][:relationships] ||= {}

      parameters[:data][:relationships]
    end
  end

  def raw_parameter_sorts
    @raw_parameter_sorts ||= parameters[:sort]
  end

  def raw_parameter_attribute_value(name)
    raw_parameter_attributes[name]
  end

  def raw_parameter_filter_value(name)
    raw_parameter_filters[name]
  end

  def raw_parameter_relationship(name)
    raw_parameter_relationships.fetch(name, {})
  end

  def raw_parameter_relationship_data(name)
    raw_parameter_relationship(name).fetch(:data, nil)
  end
end
end
end
end
# rubocop:enable Metrics/ClassLength
